#!/usr/bin/python2.6
# -*- coding: utf-8 -*-

# blipb.py - backup blip contents (including pictures)

# This script will:

#     a) backup blip.pl json archive to text files, using API
#     b) generate a script to be used later to fetch all the images
#     uploaded referenced in the archive fetched in previous step

# Just run this script and it will create "user.blip.pl-year-month.json.txt"
# files and print a shell script which used later to fetch all the images,
# to be run afterwards.

# Requirements:
# 1. a modified version of blippy.py by Patryk Bajger. My mod adds handling
#    of API's archive call and proxy support.  You can find it at
#    orfika.net.  Source can be found at http://blippy.sourceforge.net/.
# 2. Python module cjson, a very fast JSON encoder/decoder for Python by Dan
#    Pascu, available in variety of Linux distros; in Ubuntu it's
#    "python-cjson" package.  Project's homepage is at
#    http://cheeseshop.python.org/pypi/python-cjson

# Example of usage:
# $ ./blipb.py -u blipuser -p marne1234 -b rafamiga -b bando -d ./blip-backup --proxy-host=w3cache | tee run.sh ; bash run.sh

# NOTE: if you use a proxy server you have to pass it in arguments, setting
# just http_proxy environment variable doesn't help.

# NOTE: Based on blip.pl API version 0.02 to be found at http://blipapi.wikidot.com/

# Homepage: http://blipb.pl or http://orfika.net/src/blip-backup-in-python/

# Copyright 2011 Rafał Frühling <rafamiga@gmail.com>

# Revision history:
# [0.1.2] 2011-05-02
#     - initial public release
# [0.1.3] 2011-05-05
#     - (anonymous) proxy support added, to both the main script and blippy.py
# [0.1.3a] 2011-05-09
#     - fix for blindly prepending 2011 instead of current year to archive year list
# [0.1.3b] 2011-05-12
#     - a bug found! current year - 1 updates weren't fetched properly
#     - another bug found! being a non-python programmer I was using range()
#     statements wrong; meh
#     - both bugs found and fixed by Łukasz Stelmach; thanks!
# [0.2] 2011-05-31
#     - rewritten to use command-line arguments, use --help to get some info
# [0.2.1] 2011-06-24
#     - added handling of bad API credentials + some fixes of trivial stuff
# [0.2.2] -internal release-
#     - fix: setBlipArchiveUserMonth() set year, not month (sic!)
#     - fix: forgot to assign debug and verbose options
#     - add: flush stdout at several places
#     - fix: blippy.py bug when raising BadOperationException !! RE-FETCH blippy.py PLEASE !!
# [0.2.3] 2011-07-28
#     - add: -y=n/--year=n fetch only one year of data [default: current]
#     - fix: resetBlipArchive() will prevent appending data over and over and over...
# [0.2.3a] 2012-01-19
#     - fix: "no messages" message was missing an "echo"...
# To do:
#     * instead of shuffling months within one year, generate a year/mon
#       tuple, shuffle it, prepend [current] year and iterate
#     * make wget fetch multiple images at once to prevent accidental DoS
#       reports on a-little-oversesitive firewalls [like my DSL's...]

# LICENCE: GPL
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# you should have received a copy of the GNU General Public License
# along with this program (or with Nagios); if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA
#

import types
import blippy
from datetime import date
from time import localtime, strftime
import os
import sys 
import urlparse
from optparse import OptionParser
import random
import errno
import re

############ DON'T TWEAK THESE, PLEASE... ############

inmsg_pic_skel = "http://blip.pl/user_generated/update_pictures/%d_inmsg.jpg"
wget_skel = 'wget -q -nd -N -P "%s"'

blip_api_agent = "blipb.py/0.2.3"
 
############ DO NOT CHANGE BELOW THIS POINT ############

def smart_str(s, encoding='utf-8', errors='strict', from_encoding='utf-8'):
    if type(s) in (int, long, float, types.NoneType):
        return str(s)
    elif type(s) is str:
        if encoding != from_encoding:
          return s.decode(from_encoding, errors).encode(encoding, errors)
        else:
            return s

    elif type(s) is unicode:
        return s.encode(encoding, errors)
    elif hasattr(s, '__str__'):
        return smart_str(str(s), encoding, errors, from_encoding)
    elif hasattr(s, '__unicode__'):
        return smart_str(unicode(s), encoding, errors, from_encoding)
    else:
        return smart_str(str(s), encoding, errors, from_encoding)

def tstamp():
  return "[" + strftime("%a, %d %b %Y %H:%M:%S", localtime()) + "]"

class BlipMonArchive:
  def __init__(self):
    self.arch = {}
    self.picuids = []
    self.year = 0
    self.month = 0
    self.user = ""
    self.pics = {} # {picture_id: orig_url}
    
  def setupAuthenticatedAPIConnection(self):
    try:
      self.comm = blippy.Communicator(blip_api_user,blip_api_pass,blip_api_agent,use_proxy)
      self.comm.GetSubscriptions()
    except Exception, e:
      if sys.exc_type == blippy.BadCredentialsException:
        return False
      else:
        raise

    return True
      
  def resetBlipArchive(self):
    self.__init__()

  def setBlipArchiveUser(self,user = ""):
    self.user = user

  def setBlipArchiveUserYear(self,year = 0):
    if year == 0:
      self.year = date.today().year
    else:
      self.year = year

  def setBlipArchiveUserMonth(self,month = 0):
    if month == 0:
      self.month = date.today().month
    else:
      self.month = month

  def fetchArchiveData(self):
    archoffset = 0
    archlimit = 50
  
    while True:
      a = self.comm.GetArchive(self.user,self.year,self.month,archoffset,archlimit)

      if len(a) == 0:
        break
      else:
        a.reverse()
  
      for m in a:
        id = m['id']
        self.arch[m['id']] = m

        if verbose:
          mout = re.sub(r'\s', ' ', m['body_formatted'])
          print "### "+str(self.user)+" id="+str(id)+" "+m['created_at']+" "+mout.encode('utf-8','backslashreplace')
          sys.stdout.flush()

        if m['pictures_attached'] == True:
          self.picuids.append(id)

      if len(a) == archlimit:
        archoffset += archlimit
      else:
        break

    # list
    return self.arch

# {'update_path': '/updates/213366231', 'url': 'http://blip.pl/user_generated/update_pictures/1448497.jpg', 'id': 1448497}

# inmsg = http://blip.pl/user_generated/update_pictures/${id}_inmsg.jpg

  def fetchImageData(self):
    for id in self.picuids:
#      print "p id="+str(id),
      p = self.comm.GetPicture(id)[0]
      
      if verbose:
        print "### "+str(self.user)+" picid="+str(p['id'])+" url="+p['url']
        sys.stdout.flush()

      self.pics[p['id']] = p['url'];

    # dictionary - key:id value:url
    return self.pics

  def closeConnection(self):
    self.comm.closeConnection()

############# CONFIG PARSE

# some pre- stuff
yn = date.today().year
mn = date.today().month

# parse cmd parameters

try:
  p = OptionParser(version=blip_api_agent)

  # API user -- required
  p.add_option("-u", "--user", "--api-user", action="store", dest="blip_api_username", help="API username")
  # API password -- required
  p.add_option("-p", "--password", "--api-password", action="store", dest="blip_api_password", help="API password")
  # backup into directory -- required
  p.add_option("-d", "--backup-dir", action="store", dest="backup_directory", help="directory to backup into")
  # backup into directory -- required
  p.add_option("-b", "--backup-user", action="append", dest="blip_user", help="backup this user (may be used multiple times)")
  # use proxy -- optional
  p.add_option("--proxy-host", action="store", dest="proxy_hostname", help="use HTTP proxy server")
  # use proxy port -- optional
  p.add_option("--proxy-port", action="store", dest="proxy_port", default="3128", help="use proxy port (default: 3128)")
  # verbose -- optional
  p.add_option("-v", "--verbose", action="store_true", dest="verbose", default=False, help="turn on verbose output")
  # debug mode -- optional
  p.add_option("-D", "--debug", action="store_true", dest="DEBUG", default=False, help="turn on debugging messages")
  # API user -- required
  p.add_option("-y", "--year", action="store", dest="yearno", type="int", default=0, help="fetch statuses only in particular year")

  (options, args) = p.parse_args()

  # validate

  DEBUG = options.DEBUG
  verbose = options.verbose
  blip_api_user = options.blip_api_username
  blip_api_pass = options.blip_api_password
  blip_backup_dir = options.backup_directory
  backup_blip_users = options.blip_user
  yearno = 0

  if options.yearno:
      print "#### DEBUG: yearno=%d (0=disabled)" % options.yearno
      assert options.yearno > 2007, "Blip is not that old"
      assert options.yearno < yn+1, "I'm not Nostradamus"
      yearno = options.yearno

  use_proxy = ""
  if options.proxy_hostname:
    use_proxy = options.proxy_hostname + ":" + options.proxy_port
    
  assert blip_api_user != None, "API username is not set"
  assert len(blip_api_user) > 2, "API username is too short"
  assert blip_api_pass != None, "API user password is not set"
  assert len(blip_api_pass) > 6, "API user password is too short"
  assert blip_backup_dir != None, "Directory to store files in (blipb_backup_dir) is not set"
  assert len(blip_backup_dir) > 0, "Directory to store files in (blipb_backup_dir) is not set or invalid"
  assert backup_blip_users != [], "No user to backup "

except AssertionError, e:
  print >>sys.stderr, "ERROR: %s" % e
  print ""
  p.print_help()
  sys.exit(-1)

except Exception, e:
  print >>sys.stderr, "Unexpected exception=%s, %s" % (repr(e), e)
  p.print_usage()
  raise

# summary

if DEBUG:
  print "#### DEBUG: blip_api_user='%s'" % blip_api_user
  print "#### DEBUG: blip_api_pass='%s'" % blip_api_pass
  print "#### DEBUG: blip_backup_dir='%s'" % blip_backup_dir
  print "#### DEBUG: use_proxy='%s'" % use_proxy
  print "#### DEBUG: blip_user(s): [%s]" % "][".join(backup_blip_users)
  
# set location of the directories

pic_dir = blip_backup_dir + "/pics"
picfull_dir = blip_backup_dir + "/user_generated/update_pictures"
json_dir = blip_backup_dir
json_file_skel = blip_backup_dir + "/%s.blip.pl-%04d-%02d.json.txt"

if DEBUG:
  print "#### DEBUG: picfull_dir='%s'" % pic_dir
  print "#### DEBUG: picfull_dir='%s'" % picfull_dir
  print "#### DEBUG: json_dir='%s'" % json_dir
  print "#### DEBUG: json_file_skel='%s'" % json_file_skel

############# HERE WE GO

if len(blip_api_user) < 3 or len(blip_api_pass) < 6:
  raise AssertionError("blip.pl API user and/or password not set ** READ THE SCRIPT, LUKE... **")

print "# " + tstamp() + " " + blip_api_agent + " start ("+str(len(backup_blip_users))+" blip user(s))"

if use_proxy != None:
  print "# " + tstamp() + " using proxy http://" + use_proxy

# check if the credentials are right -- just fetch /subscriptions

b = BlipMonArchive()

try:
  ok = b.setupAuthenticatedAPIConnection()

except Exception, e:
  print >>sys.stderr, "\nCaught unexpected exception while setting up connection to BLIP API (%s, %s)" % (repr(e), e)
  raise

if ok == False:
  if verbose:
    print "### Cannot login to API using given user and password (user='%s')" % blip_api_user

  print >>sys.stderr, "Cannot login to API, wrong password given for user '%s'." % blip_api_user
  sys.exit(-1)

# ok, loop

print "# backup %d user(s): %s" % (len(backup_blip_users),", ".join(backup_blip_users))

if verbose:
  print "# verbose output"

print "set -x"
print "echo '" + blip_api_agent + " start'"

for user in backup_blip_users:

  if not os.path.exists(blip_backup_dir):
    print 'mkdir -p "'+blip_backup_dir+'"'

  if not os.path.exists(pic_dir):
    print 'mkdir -p "'+pic_dir+'"'

  if not os.path.exists(picfull_dir):
    print 'mkdir -p "'+picfull_dir+'"'

  try:
    os.makedirs(json_dir)
  except OSError as exc:
    if exc.errno == errno.EEXIST:
      pass
    else:
      raise
                                              
  ys = []

  if yearno == 0:
    ys = range(2007,yn)
    random.shuffle(ys)
    # fetch current year first
    ys.insert(0,yn)
  else:
    ys.insert(0,yearno)

  print "#### DEBUG: ys=%s" % str(ys)

  for y in ys:

    ms = range(1,13) # 1..12
    if y == 2007:
      ms = range(7,13) # 7..12
    if y == yn:
      ms = range(1,mn+1) # 1..now

  #  ms = [ 5 ]

    random.shuffle(ms)

    print "#### DEBUG: ms=%s" % str(ms)

    for m in ms:

      try:
        b.resetBlipArchive()
        b.setBlipArchiveUser(user)
        b.setBlipArchiveUserYear(y)
        b.setBlipArchiveUserMonth(m)

        if verbose:
          print "### u='"+user+"' y="+str(y)+" m="+str(m)

        msgs = b.fetchArchiveData()

      except Exception, e:
        if verbose:
          raise

        print >> sys.stderr, "\nAn error occured when fetching data (API call): %s (%s)" % (e, sys.exc_type)
        raise
        sys.exit(-1)
              
      msgsc = len(msgs)

      fn = json_file_skel % (user,y,m)
      print "# " + tstamp() + " " + user + " blip#("+str(y)+"," + "%02d" % m + ") msgs="+str(msgsc)+ " fn="+fn
      
      if len(msgs) > 0:
      
        try:
          f = open(fn, 'w')
          try:
            f.write(str(msgs))
          finally:
            f.close()

        except IOError as (errno, strerror):
          print "I/O error({0}): {1}".format(errno, strerror),

        except:
          print "Unexpected error: ", sys.exc_info()[0]
          raise

        pics = b.fetchImageData()
        picsc = len(pics)

        print "# " + tstamp() + " " + user + " blip#("+str(y)+"," + "%02d" % m + ") pics#=" + str(picsc)

        b.closeConnection()

        print "echo '" + user + " " + str(y) + "/" + "%02d" % m + " has " + str(picsc) + " pic(s)'"

        print "# " + tstamp() + " " + user + " fetch " + str(picsc) + " pic(s)"

        sys.stdout.flush()

        for id,url in pics.iteritems():
#          print 'wget -q -P "'+pic_dir+'/'+str(id)+'_inmsg.jpg" '''+inmsg_pic_skel%id+"'"
          if use_proxy != None:
            print "http_proxy=http://" + use_proxy,
          print wget_skel % pic_dir + " '"+inmsg_pic_skel%id+"'"

          if use_proxy != None:
            print "http_proxy=http://" + use_proxy,
          print wget_skel % picfull_dir + " '"+url+"'"

      else:
        print "echo '" + str(y) + "/" + "%02d" % m + " was a lazy for " + user + ", Blip reports no messages in that month...'"

      sys.stdout.flush()
        
print "# " + tstamp() + " meta"
print "echo 'meta " + blip_api_agent + "'"
